#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Isaac: Interface for Scanner to Automate Acquisition on Client

Copyright (C) 2003, 2004 - Marco Andreini, Cristian Lucchesi

GPL 2.1

"""

import sys
import os
import threading
import logging

import twain
import wx

from urlparse import urlparse, urlunparse
from httppost import HTTPPost

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# py >= 2.3 | optik
from optparse import OptionParser

# subclassing StringIO
from StringIO import StringIO as StringIOpy

# wxImage don't save to .png (20/02/2004)
import Image
import PngImagePlugin
import BmpImagePlugin
# don't look for more plugin
Image._initialized = 1


__version__ = "$Id$"

version = "0.6"


class IsaacClientException(Exception):
    """internal isaac-client error"""
    
    pass


class MainFrame(wx.Frame):
    """main frame"""

    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, wx.DefaultPosition,
                          wx.Size(400,300))
        #b = wxButton(self, 20, "Acquire", wxPoint(20, 60), wxSize(100, 45))
        #EVT_BUTTON(self, 20, self.OnClick)
        #b.SetToolTipString("Start acquisition...")
        
        self.CreateStatusBar()
        menu = wx.Menu()

        idOpenScanner = wx.NewId()
        idAcquire = wx.NewId()
        idExit = wx.NewId()
        idScrolledWindow1 = wx.NewId()
        idBMPImage = wx.NewId()
        
        # self.GetClientSize().x , ...y
        
        self.scrolledWindow1 = wx.ScrolledWindow(id=idScrolledWindow1,
                                                name="scrolledWindow1",
                                                parent=self,
                                                pos=wx.Point(0, 0),
                                                size=self.GetClientSize(),
                                                style=wx.TAB_TRAVERSAL | wx.SUNKEN_BORDER)
        self.text = wx.StaticText(self.scrolledWindow1, -1, "Acquiring...", (50, 50))
        self.SM, self.SD = (None, None)
        
    #def OnClick(self, event):
    #    self.MnuAcquire(event)
    
    #def MnuQuit(self, event):
    #    self.Close(1)
    
    def OnClose(self, event):
        # There seems to be some problems closing down, if the
        # twain objects are not destroyed correctly. Therefore
        # trap the window close event.
        if self.SD:
            self.SD.destroy()
        if self.SM:
            self.SM.destroy()
        del self.SD, self.SM
        self.sender.stop()
        self.Destroy()
    
    def OpenScanner(self, event):
        if not self.SM:
            self.SM = twain.SourceManager(self.GetHandle(), ProductName="IsAAC Acquisitor")
        if not self.SM:
            return
        
        self.SM.SetCallback(self.OnTwainEvent)
        
        scannerList = self.SM.GetSourceList()
        if len(scannerList) > 1:
            # FIXME: choose the scanner
            logger.warn("more scanners present, choose first one.")
        src = scannerList[0]
        self.SD = self.SM.OpenSource(src)
        if self.SD:
            self.SetTitle("Acquisitor - %s" % (self.SD.GetSourceName(), ))
    
    def RunAcquire(self, event=None):
        if not self.SD:
            self.OpenScanner(event)
        if not self.SD:
            return
        
        # FIXME: set others requested capabilities
        # Scale image to 100%
        self.SD.SetCapability(twain.ICAP_YRESOLUTION, twain.TWTY_FIX32, 100.0)
        # with no args (default is True) show user interface
        self.SD.RequestAcquire(False)
    
    def ProcessTransfer(self, event):
        moreToCome = 1
        while moreToCome:
            (handle, moreToCome) = self.SD.XferImageNatively()
            self.sender.addFile(twain.DIBToBMFile(handle))
            twain.GlobalHandleFree(handle)
        self.sender.finish()
    
    def OnTwainEvent(self, event):
        if event == twain.MSG_XFERREADY:
            self.ProcessTransfer(event)
            self.SD = None
        elif event == twain.MSG_CLOSEDSREQ:
            self.SD = None

    def NotifyMsg(self, msg):
        self.text.SetLabel(msg)
        self.Refresh()
        wx.Yield()


class IsaacClientBase:
    """isaac client (no threaded)"""
    
    destImageType = "png"
    
    def __init__(self, filename, endFunction, notifyFunction=None):
        self.endFunction = endFunction
        self.notifyFunction = notifyFunction
        
        inputFile = open(filename, "rb")
        metadata = {}
        
        for l in inputFile.readlines():
            key, value = l.split(":", 1)
            key = key.strip().lower()
            value = value.strip()
            metadata[key] = value
        
        # parse incoming url
        schema, host, path = urlparse(metadata["url"])[:3]

        # WARNING: string comparation.
        if metadata["version"] > version:
            # FIXME: windowed message?
            raise IsaacClientException, "version %r required" % (metadata["version"], )

        url = urlunparse((schema, host, path, "", "", ""))
        self._request = HTTPPost(url)
        self._request.addCookie(metadata.get("cookie"))
        
        # only for notify/debug
        self.counter = 1

    def start(self):
        """fake start"""
        
        pass
        
    def addFile(self, fileData):
        """addFile(file_content)
        
        Note: is safe to call from others threads
        """

        logger.debug("addFile: start")
        # problem with twain image-handle? passing file-data to StringIO
        infile = StringIO(fileData)

        logger.debug("addFile: notify(processing...")
        self.notify("processing image %d" % self.counter)

        # without PIL:
        #im = wxImageFromStream(infile)

        #outstream =?
        #im.saveToPNG(outstream)
        
        #fname = tempfile.mktemp()
        #ok, msg = img2img.convert(image_file, maskClr, None, tfname, wx.wxBITMAP_TYPE_PNG, ".png")
        #if not ok:
        #    print msg
        #    return
        #data = open(tfname, "rb").read()
        #data = crunch_data(data, compressed)
        #os.unlink(tfname)

        ##with PIL:
        logger.debug("addFile: image.open")
        im = Image.open(infile)
        
        logger.debug("addFile: stringIO")
        outfile = StringIO()

        logger.debug("addFile: Image.open, im.save")
        # only-if im.format.lower() != self.destImageType
        im.save(outfile, self.destImageType)
        #outfile.reset()

        logger.debug("addFile: request.add_file")
        self._request.addFile("file:list",
                              "image-%d.png" % (self.counter, ),
                              "image/%s" % (self.destImageType, ),
                              outfile.getvalue())
        self.counter += 1
        logger.debug("addFile: end")
    
    def stop(self):
        """close thread"""
        
        self.notify("stop requested in non-threaded client!")
        
    def finish(self):
        """end & send data to server"""
        
        self.notify("sending data to server...")
        code, message, headers = self._request.getreply()
        if code != 200:
            # TODO: msgbox?
            raise IsaacClientException, "trasmission/server error (%r, %r)" % \
                  (code, message)
        
        self.endFunction()
        
    def notify(self, *args, **kwargs):
        """notify(*args, **kwargs)"""
        
        if not self.notifyFunction is None:
            self.notifyFunction(*args, **kwargs)


class IsaacClientThreaded(IsaacClientBase, threading.Thread):
    """isaac client"""
    
    def __init__(self, filename, endFunction, notifyFunction=None):
        IsaacClientBase.__init__(self, filename, endFunction, notifyFunction)
        
        # FIFO list with locking
        self._files = []
        self._filesLock = threading.Lock()
        
        # to locking this thread, waiting if self._files is empty
        self._semaphore = threading.Semaphore(0)
        # end condition
        self._stopRequested = False
        self._finish = False
        
        threading.Thread.__init__(self, name="IsaacClient")
    
    def addFile(self, file):
        """addFile(file_like_object)
        is safe to call from others threads"""
        
        # problem with twain image-handle? passing file-data to StringIO
        self._filesLock.acquire()
        self._files.append(StringIO(file))
        # unlock this thread, if locked
        self._semaphore.release()
        self._filesLock.release()
    
    def stop(self):
        """close thread"""
        
        self._stopRequested = True
        self._semaphore.release()

    def finish(self):
        """end & send data to server"""
        
        self._stopRequested = True
        self._finish = True
        self._semaphore.release()
        
    def run(self):
        """acquisitor-loop"""
        
        # only for message
        counter = 1
        
        while 1:
            self._semaphore.acquire()
            # NOTE: without locking _files
            if self._stopRequested and not self._files:
                break
            
            self._filesLock.acquire()
            # pop first value
            file = self._files.pop(0)
            self._filesLock.release()
            
            self.notify("processing image %d" % counter)
            
            # convert to imageType in outfile
            
            im = Image.open(file)
            outfile = StringIO()
            # only-if im.format.lower() != self.destImageType
            im.save(outfile, self.destImageType)
            outfile.reset()
            
            self._request.addFile("file:list",
                                  "image-%d" % (counter, ),
                                  "image/%s" % (self.destImageType, ),
                                  outfile.getvalue())
            counter += 1
        
        if not self._finish:
            raise IsaacClientException, "stop requested..."
        
        self.notify("sending data to server...")
        code, message, headers = self._request.getreply()
        if code != 200 or code != 302:
            # TODO: msgbox?
            raise IsaacClientException, "trasmission/server error (%r, %r)" % \
                  (code, headers)
        
        self.endFunction()


class Acquisitor(wx.App):
    """acquisitor client"""

    def OnInit(self):
        """oninit"""
        
        self.frame = frame = MainFrame(None, -1, "Acquisitor")
        frame.Show(True)
        self.SetTopWindow(frame)

        parser = OptionParser(version=__version__)
        parser.add_option("-t", "--threaded",
                          action="store_true", dest="threaded",
                          help="using thread for main logic (fast mode?)",
                          default=False)
        parser.add_option("-v", "--verbose", default=False,
                          action="store_true", dest="verbose",
                          help="print status messages to stderr")
        parser.add_option("-n", "--no-log", dest="logfile",
                          action="store_false",
                          help="no log file generated, debug on stderr",
                          default=True)
    
        (options, args) = parser.parse_args()

        if options.verbose:
            logger.setLevel(logging.DEBUG)
        
        if options.logfile:
            # debug.log got to python/isaacAcquisitor.exe binary directory
            # or script directory
            # or current directory
            # (or maybe to temp?)
            # or to stderr.
            for logDirectory in [os.path.dirname(sys.executable),
                                 os.path.dirname(sys.argv[0]),
                                 os.getcwd()]:
                debugFileName = os.path.join(logDirectory, "isaac.log")
                try:
                    sys.stderr = open(debugFileName, "a")
                except IOError, err:
                    # WARNING: assume permission denied
                    continue
                break
            # file formatted logger
            handler = logging.StreamHandler(sys.stderr)
            formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            
            # start of work
            logger.info("started")
        
        if len(args) < 1:
            # do this via stderr?
            dlg = wx.MessageDialog(frame, "Manca il parametro \"nomefile\"", "Errore",
                                   wx.OK | wx.ICON_ERROR)
                                  #wxYES_NO | wxNO_DEFAULT | wxCANCEL | wxICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
            raise IsaacClientException, "parametro \"nomefile\" richiesto"

        filename = args[0]
        if options.threaded:
            IsaacClient = IsaacClientThreaded
        else:
            IsaacClient = IsaacClientBase
        self.sender = IsaacClient(filename, frame.Close, frame.NotifyMsg)
        self.sender.start()

        frame.sender = self.sender
        frame.RunAcquire()
        return True
    
    # WARNING: called?
    def OnClose(self):
        self.sender.stop()


if __name__ == "__main__":
    # create generic logger
    logger = logging.getLogger("isaac")
    logger.setLevel(logging.WARNING)
    aq = Acquisitor(0)
    aq.MainLoop()
